from typing import Callable, Any
from functools import wraps
from dataclasses import dataclass


@dataclass
class DependentDefault:
    dep: Callable


def updatable_default(updator: Callable = lambda x, y: x | y, **defaults):
    """
    Ever wished that instead of *overriding* a function's default
    parameters with those supplied in your call, you could just
    "update" it with that supplied in your call? Something like
    f(y += 1)?

    You can do this with `updatable_default`. Instead of defining the default
    parameter of the function in the standard way, do it like this:

    @updatable_default(x = {"a": 1})
    def f(x):
        return x

    f() # returns {"a": 1}
    f(x = {"b": 2}) # returns {"a": 1, "b": 2}
    f(x = {"a": 3}) # returns {"a": 3}

    @updatable_default(updator = lambda x, y: x + y, y = 8)
    def f(y):
        return y

    f(3) # returns 11

    def update_recursive(source, overrides):
        for key, value in overrides.items():
            if isinstance(value, dict) and key in source and isinstance(source[key], dict):
                update_recursive(source[key], value)
            elif isinstance(value, list) and key in source and isinstance(source[key], list):
                source[key] = source[key] + value
            else:
                source[key] = value
        return source


    @updatable_default(
        post={
            "content": "Hello, World!",
            "OP": {
                "name": "John",
                "age": 30,
                "metadata": {
                    "account_created": "2024-01-01",
                    "banned": False,
                    "votes_cast": 1,
                },
            },
            "metadata": {"score": -2, "flags": ["spam"]},
        },
        updator=update_recursive,
    )
    def f(post):
        return post


    f(
        post={
            "OP": {
                "metadata": {"account_deleted": "2024-01-02", "banned": True},
            },
            "metadata": {"flags": ["deleted"]},
        },
    )

    ### returns:
    {
        "content": "Hello, World!",
        "OP": {
            "name": "John",
            "age": 30,
            "metadata": {
                "account_created": "2024-01-01",
                "banned": True,
                "votes_cast": 1,
                "account_deleted": "2024-01-02",
            },
        },
        "metadata": {"score": -2, "flags": ["spam", "deleted"]},
    }

    Stuff this can be useful for:

    - tracing
    - cumulative configurations
    - "biasing" or "damping" some kind of computation

    The default updator is `|`.

    You can also have a default be dependent on func, args and kwargs by
    setting it to an object DependentDefault(dep=lambda ...)
    """

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for arg_name, default_value in defaults.items():
                if isinstance(default_value, DependentDefault):
                    default_value = default_value.dep(func, *args, **kwargs)
                if arg_name in kwargs:
                    kwargs[arg_name] = updator(default_value, kwargs[arg_name])
                # else:
                #     import inspect

                #     sig = inspect.signature(func)
                #     param_position = list(sig.parameters.keys()).index(arg_name)
                #     if param_position < len(args):
                #         args = list(args)
                #         args[param_position] = updator(
                #             default_value, args[param_position]
                #         )
                #         args = tuple(args)
                else:  # if arg_name not provided, add it and set it to default
                    kwargs[arg_name] = default_value
            return func(*args, **kwargs)

        return wrapper

    return decorator


def invoice(description: str = None):
    """
    Example usage:

    @invoice()
    def manufacture_cups(**kwargs):
        manufacture_steel(**kwargs)
        educate_potters(**kwargs)
        return

    @invoice("educating potters")
    def educate_potters(**kwargs):
        capture_enemies(**kwargs)
        return

    @invoice("manufacturing steel")
    def manufacture_steel(**kwargs):
        mine_iron(**kwargs)
        return

    @invoice("mining iron")
    def mine_iron(**kwargs):
        print(kwargs["description"])
        return

    @invoice("capturing enemies")
    def capture_enemies(**kwargs):
        print(kwargs["description"])
        return

    manufacture_cups()
    ### prints:
    [{'func': 'manufacture_cups', 'args': [], 'kwargs': {}}, 'manufacturing steel', 'mining iron']
    [{'func': 'manufacture_cups', 'args': [], 'kwargs': {}}, 'educating potters', 'capturing enemies']

    If description is not provided, it will default to the function's name
    and arguments like above.

    """
    if description is None:
        description = DependentDefault(
            dep=lambda func, *args, **kwargs: {
                "func": func.__name__,
                "args": list(args),
                "kwargs": kwargs,
            }
        )
    return updatable_default(
        description=description,
        updator=lambda default, new: (
            new + [default] if isinstance(new, list) else [new] + [default]
        ),
    )


import asyncio
from functools import wraps
from dataclasses import dataclass
from copy import deepcopy
from typing import Callable, Any
from costly.costlog import Costlog
from costly.simulators.llm_simulator_faker import LLM_Simulator_Faker
from costly.estimators.llm_api_estimation import LLM_API_Estimation
from inspect import signature, Parameter, iscoroutinefunction


@dataclass
class CostlyResponse:
    output: Any
    cost_info: dict[str, Any]


def costly(
    simulator: Callable = LLM_Simulator_Faker.simulate_llm_call,
    estimator: Callable = LLM_API_Estimation.get_cost_real,
    **param_mappings: dict[str, Callable]
):
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            cost_log = kwargs.pop("cost_log", None)
            simulate = kwargs.pop("simulate", False)
            description = kwargs.pop("description", None)
            
            # Get the function's signature
            sig = signature(func)
            
            # Create a dictionary with default values
            options = {
                k: v.default for k, v in sig.parameters.items()
                if v.default is not Parameter.empty
            }

            # Update default kwargs with provided kwargs
            options.update(kwargs)

            # apply param_mappings
            costly_kwargs = options | {
                key: mapping(options) if callable(mapping) else options.get(mapping)
                for key, mapping in param_mappings.items()
            }

            if simulate:
                simulator_kwargs = {
                    k: v
                    for k, v in costly_kwargs.items()
                    if k in signature(simulator).parameters
                } | {"cost_log": cost_log, "description": description}
                return simulator(**simulator_kwargs)

            if cost_log is not None:
                async with cost_log.new_item_async() as (item, timer):
                    output = await func(*args, **kwargs)  # await the coroutine
                    cost_info = {}
                    if isinstance(output, CostlyResponse):
                        output, cost_info = output.output, output.cost_info
                    estimator_kwargs = (
                        {
                            k: v
                            for k, v in costly_kwargs.items()
                            if k in signature(estimator).parameters
                        }
                        | {
                            "output_string": output,
                            "description": description,
                            "timer": timer(),
                        }
                        | cost_info
                    )
                    cost_item = estimator(**estimator_kwargs)
                    item.update(cost_item)
            else:
                output = await func(*args, **kwargs)
                if isinstance(output, CostlyResponse):
                    output, cost_info = output.output, output.cost_info
            return output

        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            cost_log = kwargs.pop("cost_log", None)
            simulate = kwargs.pop("simulate", False)
            description = kwargs.pop("description", None)
            
            # Get the function's signature
            sig = signature(func)
            
            # Create a dictionary with default values
            options = {
                k: v.default for k, v in sig.parameters.items()
                if v.default is not Parameter.empty
            }

            # Update default kwargs with provided kwargs
            options.update(kwargs)

            # apply param_mappings
            costly_kwargs = options | {
                key: mapping(options) if callable(mapping) else options.get(mapping)
                for key, mapping in param_mappings.items()
            }

            if simulate:
                simulator_kwargs = {
                    k: v
                    for k, v in costly_kwargs.items()
                    if k in signature(simulator).parameters
                } | {"cost_log": cost_log, "description": description}
                return simulator(**simulator_kwargs)

            if cost_log is not None:
                with cost_log.new_item() as (item, timer):
                    output = func(*args, **kwargs)  # call function normally
                    cost_info = {}
                    if isinstance(output, CostlyResponse):
                        output, cost_info = output.output, output.cost_info
                    estimator_kwargs = (
                        {
                            k: v
                            for k, v in costly_kwargs.items()
                            if k in signature(estimator).parameters
                        }
                        | {
                            "output_string": output,
                            "description": description,
                            "timer": timer(),
                        }
                        | cost_info
                    )
                    cost_item = estimator(**estimator_kwargs)
                    item.update(cost_item)
            else:
                output = func(*args, **kwargs)
                if isinstance(output, CostlyResponse):
                    output, cost_info = output.output, output.cost_info
            return output

        return async_wrapper if iscoroutinefunction(func) else sync_wrapper

    return decorator



import types
import typing
import random
from typing import Any
from faker import Faker
from pydantic import BaseModel
from costly.costlog import Costlog
from costly.estimators.llm_api_estimation import LLM_API_Estimation


class LLM_Simulator_Faker:
    """
    Main things you would likely subclass:
    - FAKER_PARAMS: default is:
        ```python
        {
            "str": {
                "max_nb_chars": 600 * 4.5,
            },
            "int": {
                "min": 0,
                "max": 100,
            },
            "float": {
                "a": -20,
                "b": 20,
            },
        }
        ```
        Override this to make it generate different lengths of text, etc.
    - _fake_custom: to add highest-priority/overriding behaviour for your own types
    - tryfuncs: to index any new faking methods you define or change priorities
    - redirects: to change type redirects like list -> list[str], object -> str, etc.

    Example:

    ```python
    class MySimulator(LLM_Simulator_Faker):

        tryfuncs = super().tryfuncs + [
            {"func": _fake_pandas_df, "priority": 100},
        ]

        def _fake_pandas_df(self, t: type):
            if t.__module__ == "pandas" and t.__name__ == "DataFrame":
                return pd.DataFrame(...)
            else:
                raise ValueError(f"Unsupported type: {t}")

        def _fake_custom(self, t: type):
            if t == MyType:
                ...
            else:
                raise ValueError(f"Unsupported type: {t}")
    ```

    Attributes:
        FAKER: Faker instance
        tryfuncs: list of tryfuncs with priorities
        redirects: dict of types to redirect to

    Methods:
        simulate_llm_call: simulate an LLM call
        fake: fake a value of a given type
        _fake_basic: fake a basic value
        _fake_paramgeneric: fake a value of a generic type
        _fake_redirect: fake a value of a redirected type
        _fake_pydantic: fake a value of a pydantic type
        _fake_datetime: fake a value of a datetime type
        _fake_uuid: fake a value of a uuid type
        _fake_typing: fake a value of a typing type
        _fake_custom: fake a value of a custom type
    """

    FAKER = Faker()

    FAKER_PARAMS = {
        "str": {
            "max_nb_chars": 600 * 4.5,
        },
        "int": {
            "min": 0,
            "max": 100,
        },
        "float": {
            "a": -20,
            "b": 20,
        },
        "size_set": {
            "min": 0,
            "max": 10,
        },
        "size_dict": {
            "min": 0,
            "max": 10,
        },
        "size_list": {
            "min": 0,
            "max": 10,
        },
    }

    @classmethod
    def simulate_llm_call(
        cls,
        input_string: str = None,  # must supply at least one of input_string, input_tokens, messages
        input_tokens: int = None,
        messages: list[dict[str, str]] = None,
        model: str = None,
        response_model: type = str,
        cost_log: Costlog = None,
        description: list[str] = None,
    ) -> str | Any:
        """
        Simulate an LLM call.
        """
        response_model = response_model or str
        response = cls.fake(response_model)
        if cost_log is not None:
            assert model is not None, "model is required for tracking costs"
            with cost_log.new_item() as (item, _):
                cost_item = LLM_API_Estimation.get_cost_simulating(
                    input_string=input_string,
                    input_tokens=input_tokens,
                    model=model,
                    messages=messages,
                    description=description,
                    # output_string=response, # not needed
                )
                item.update(cost_item)
        return response

    @classmethod
    def simulate_llm_probs(
        cls,
        return_probs_for: list[str],
        input_string: str = None,  # must supply at least one of input_string, input_tokens, messages
        input_tokens: int = None,
        messages: list[dict[str, str]] = None,
        model: str = None,
        cost_log: Costlog = None,
        description: list[str] = None,
    ) -> dict[str, float]:
        """
        Simulate a dict with keys as return_probs_for and values as probabilities.
        """
        probs_ = {key: random.random() for key in return_probs_for}
        total = sum(probs_.values())
        probs = {key: value / total for key, value in probs_.items()}
        if cost_log is not None:
            assert model is not None, "model is required for tracking costs"
            with cost_log.new_item() as (item, _):
                cost_item = LLM_API_Estimation.get_cost_simulating(
                    input_string=input_string,
                    input_tokens=input_tokens,
                    output_tokens_min=0,
                    output_tokens_max=max(len(key) for key in return_probs_for),
                    model=model,
                    messages=messages,
                    description=description,
                )
                item.update(cost_item)
        return probs

    @classmethod
    def fake(cls, t: type):
        try_funcs_in_order = sorted(
            cls.tryfuncs, key=lambda x: x["priority"], reverse=True
        )
        exceptions = []
        for try_func in try_funcs_in_order:
            try:
                return try_func["func"](cls, t)
            except Exception as e:
                exceptions.append({"try_func": try_func, "exception": e})
        raise ValueError(
            f"Unsupported type: {t}\n"
            + "\n".join([str(e) for e in exceptions])
        )

    @classmethod
    def _fake_basic(cls, t: type):
        if t == type(None):
            return None
        elif isinstance(t, types.UnionType):
            t = cls.FAKER.random_element(elements=t.__args__)
            return cls.fake(t)
        elif t == str:
            return cls.FAKER.text(**cls.FAKER_PARAMS["str"])
        elif t == int:
            return cls.FAKER.random_int(**cls.FAKER_PARAMS["int"])
        elif t == float:
            return random.uniform(**cls.FAKER_PARAMS["float"])
        elif t == bool:
            return cls.FAKER.random_element(elements=[True, False])
        else:
            raise ValueError(f"Unsupported type: {t}")

    @classmethod
    def _fake_paramgeneric(cls, t: type):
        origin, args = t.__origin__, t.__args__
        if origin == list:
            return [
                cls.fake(args[0])
                for _ in range(cls.FAKER.random_int(**cls.FAKER_PARAMS["size_list"]))
            ]
        elif origin == dict:
            return {
                cls.fake(args[0]): cls.fake(args[1])
                for _ in range(cls.FAKER.random_int(**cls.FAKER_PARAMS["size_dict"]))
            }
        elif origin == tuple:
            return tuple(cls.fake(args[i]) for i in range(len(args)))
        elif origin == set:
            return {
                cls.fake(args[0])
                for _ in range(cls.FAKER.random_int(**cls.FAKER_PARAMS["size_set"]))
            }
        else:
            raise ValueError(f"Unsupported type: {t}")

    @classmethod
    def _fake_redirect(cls, t: type):
        if t in cls.redirects:
            return cls.fake(cls.redirects[t])
        else:
            raise ValueError(f"Unsupported type: {t}")

    @classmethod
    def _fake_pydantic(cls, t: type):
        assert issubclass(t, BaseModel)
        factory_dict = {}
        for name, field in t.model_fields.items():
            factory_dict[name] = cls.fake(field.annotation)
        return t(**factory_dict)

    @classmethod
    def _fake_datetime(cls, t: type):
        assert t.__module__ == "datetime"
        if t.__name__ == "datetime":
            return cls.FAKER.date_time_this_decade()
        elif t.__name__ == "date":
            return cls.FAKER.date_this_decade()
        elif t.__name__ == "time":
            return cls.FAKER.time_object()
        else:
            raise ValueError(f"Unsupported type: {t}")

    @classmethod
    def _fake_uuid(cls, t: type):
        assert t.__module__ == "uuid" and t.__name__ == "UUID"
        return cls.FAKER.uuid4(cast_to=None)

    @classmethod
    def _fake_typing(cls, t: type):
        assert t.__module__ == "typing"
        if t.__name__ == "Any":
            t = object
        elif t.__name__ == "Union":
            t = cls.FAKER.random_element(elements=t.__args__)
        elif t.__name__ == "Optional":
            t = cls.FAKER.random_element(elements=[type(None), t.__args__[0]])
        elif typing.get_origin(t) is not None:
            t = typing.get_origin(t)
        else:
            raise ValueError(f"Unsupported type: {t}")
        return cls.fake(t)

    @classmethod
    def _fake_custom(cls, t: type):
        raise ValueError("_fake_custom must be subclassed to be used")

    tryfuncs = [
        {"func": lambda cls, t: cls._fake_custom(t), "priority": 1000},
        {"func": lambda cls, t: cls._fake_paramgeneric(t), "priority": 500},
        {"func": lambda cls, t: cls._fake_pydantic(t), "priority": 100},
        {"func": lambda cls, t: cls._fake_basic(t), "priority": 100},
        {"func": lambda cls, t: cls._fake_datetime(t), "priority": 100},
        {"func": lambda cls, t: cls._fake_uuid(t), "priority": 100},
        {"func": lambda cls, t: cls._fake_typing(t), "priority": 100},
        {"func": lambda cls, t: cls._fake_redirect(t), "priority": -100},
    ]
    """
    Priority of tryfuncs:
    - _fake_custom: 1000
    - _fake_paramgeneric: 500
    - _fake_pydantic: 100
    - _fake_basic: 100
    - _fake_datetime: 100
    - _fake_uuid: 100
    - _fake_typing: 0
    - _fake_redirect: -100
    
    Important that:
    - _fake_custom is highest, as it should be used to override
      all other behaviour
    - _fake_redirect is lowest, to avoid unwanted redirects e.g. for parametric
      generics
    - _fake_typing is also pretty low, so that _fake_paramgeneric
      can do its job when possible
    """

    redirects = {
        list: list[str],
        tuple: tuple[str, str],
        set: set[str],
        dict: dict[str, str],
        object: str,
    }



import tiktoken
import logging
from io import StringIO
from pydantic import BaseModel
import ast
import warnings
from unittest.mock import patch


class LLM_API_Estimation:
    """Library of functions for creating cost items from LLM API calls.
    You may subclass this and override any of the attributes or methods.

    Attributes:
    - PRICES: dictionary of prices for each model

    Methods to produce cost items:
    - get_cost_simulating(...) -> dict[str, float]
    - get_cost_real(...) -> dict[str, float]

    Auxillary methods:
    - get_model(model, supported_models=PRICES.keys()) -> model (get the model with the longest prefix matching model)
    - get_prices(model, price_dict=PRICES) -> prices
    - tokenize(text, model) -> input_tokens (tokenize text with tiktoken)
    - output_tokens_estimate(input_tokens, model) -> output_tokens_min, output_tokens_max
      (estimate output tokens for simulating)

    Methods to help convert things to input_tokens or input_string:
    - messages_to_input_tokens(messages: list[dict[str, str]]) -> int
    - prompt_to_input_tokens(prompt: str, system_prompt: str = None, model: str = None) -> int
    - get_raw_input_string_instructor(messages: str | list[dict[str, str]], response_model:BaseModel) -> str

    Private methods:
    - _get_cost_simulating_from_input_tokens_output_tokens(...) -> dict[str, float]
    - _get_cost_real_from_input_tokens_output_tokens_timer(...) -> dict[str, float]
    - _get_tokens(...) -> tuple[int, int, int]
    - _process_raw_prompt(...) -> str
    - _tokenize_rough(text, model) -> input_tokens

    Subclassing tips:
    - PRICES: to change the prices/times for each model, or support new models or model names e.g.
      PRICES = LLM_API_Estimation.PRICES | {"my_model": LLM_API_Estimation.PRICES["gpt-4o"]}
    - tokenize: e.g.
      tokenize=_tokenize_rough
    - output_tokens_estimate: to change the way you estimate output tokens for simulating
    - get_model: e.g. if you want to just match models literally --
      get_model=lambda model, supported_models: model
    - get_raw_input_string_instructor: e.g. if instructor adds a better way to see the raw prompt
    - _process_raw_prompt: e.g. if instructor adds a better way to see the raw prompt
    - _get_cost_simulating_from_input_tokens_output_tokens,
      _get_cost_real_from_input_tokens_output_tokens_timer,
      get_cost_simulating,
      get_cost_real
      if you're going to change these (e.g. to use this library for some purpose other than
      LLM API calls), maybe just define a new class
    """

    PRICES = {
        "gpt-4o": {
            "input_tokens": 5.0e-6,
            "output_tokens": 15.0e-6,
            "time": 18e-3,
        },
        "gpt-4o-mini": {
            "input_tokens": 0.15e-6,
            "output_tokens": 0.6e-6,
            "time": 9e-3,
        },
        "gpt-4-turbo": {
            "input_tokens": 10.0e-6,
            "output_tokens": 30.0e-6,
            "time": 36e-3,
        },
        "gpt-4": {
            "input_tokens": 30.0e-6,
            "output_tokens": 60.0e-6,
            "time": 36e-3,
        },
        "gpt-3.5-turbo": {
            "input_tokens": 0.50e-6,
            "output_tokens": 1.50e-6,
            "time": 36e-3,
        },
        "claude-3-5-sonnet": {
            "input_tokens": 3.0e-6,
            "output_tokens": 15.0e-6,
            "time": 18e-3,
        },
        "claude-3-opus": {
            "input_tokens": 15.0e-6,
            "output_tokens": 75.0e-6,
            "time": 18e-3,
        },
        "claude-3-haiku": {
            "input_tokens": 0.25e-6,
            "output_tokens": 1.25e-6,
            "time": 9e-3,
        },
    }
    """
    input_tokens: dollar per input token
    output_tokens: dollar per output token
    time: seconds per output token
    """

    @classmethod
    def get_model(cls, model: str, supported_models: list = None):
        """Get model in supported_models with the longest prefix matching model"""
        if supported_models is None:
            supported_models = cls.PRICES.keys()

        matching_keys = [key for key in supported_models if model.startswith(key)]
        if not matching_keys:
            raise ValueError(f"No matching model found for {model}")

        return max(matching_keys, key=len)

    @classmethod
    def get_prices(cls, model: str, price_dict: dict = None):
        """Get prices for a model"""
        if price_dict is None:
            price_dict = cls.PRICES

        return price_dict[cls.get_model(model, price_dict.keys())]

    @classmethod
    def tokenize(cls, input_string: str, model: str) -> int:
        """Tokenize text"""
        supported_models = [
            "gpt-4",
            "gpt-3.5-turbo",
            "text-embedding-ada-002",
            "text-embedding-3-small",
            "text-embedding-3-large",
        ]
        try:
            encoding = tiktoken.encoding_for_model(
                cls.get_model(model, supported_models)
            )
        except:
            encoding = tiktoken.get_encoding("cl100k_base")
        return len(encoding.encode(input_string))

    @classmethod
    def _tokenize_rough(cls, input_string: str, model: str = None) -> int:
        "For a quick estimate, just divide by 4.5"
        return len(input_string) // 4.5

    @classmethod
    def output_tokens_estimate(
        cls,
        input_string: str = None,
        messages: list[dict[str, str]] = None,
        input_tokens: int = None,
        model: str = None,
    ) -> tuple[int, int]:
        return [0, 2048]

    @classmethod
    def messages_to_input_tokens(
        cls,
        messages: list[dict[str, str]], model: str = None
    ) -> int:
        """Return the number of tokens used by a list of messages.

        From: https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb
        """
        try:
            encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            warnings.warn(
                "messages_to_input_tokens: model not found. Using cl100k_base encoding."
            )
            encoding = tiktoken.get_encoding("cl100k_base")
        if model in {
            "gpt-3.5-turbo-0613",
            "gpt-3.5-turbo-16k-0613",
            "gpt-4-0314",
            "gpt-4-32k-0314",
            "gpt-4-0613",
            "gpt-4-32k-0613",
        }:
            tokens_per_message = 3
            tokens_per_name = 1
        elif model == "gpt-3.5-turbo-0301":
            tokens_per_message = (
                4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
            )
            tokens_per_name = -1  # if there's a name, the role is omitted
        elif "gpt-3.5-turbo" in model:
            warnings.warn(
                "messages_to_input_tokens: gpt-3.5-turbo may update over time. Returning num tokens assuming gpt-3.5-turbo-0613."
            )
            return cls.messages_to_input_tokens(
                messages, model="gpt-3.5-turbo-0613"
            )
        elif "gpt-4" in model:
            warnings.warn(
                "messages_to_input_tokens: gpt-4 may update over time. Returning num tokens assuming gpt-4-0613."
            )
            return cls.messages_to_input_tokens(
                messages, model="gpt-4-0613"
            )
        else:
            warnings.warn(
                f"messages_to_input_tokens: model {model} not found. Returning num tokens assuming gpt-4-0613."
            )
            return cls.messages_to_input_tokens(
                messages, model="gpt-4-0613"
            )
        num_tokens = 0
        for message in messages:
            num_tokens += tokens_per_message
            for key, value in message.items():
                num_tokens += len(encoding.encode(value))
                if key == "name":
                    num_tokens += tokens_per_name
        num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
        return num_tokens

    @classmethod
    def prompt_to_messages(
        cls,
        prompt: str,
        system_prompt: str = None,
    ) -> list[dict[str, str]]:
        messages = []
        if system_prompt is not None:
            messages.append({"content": system_prompt, "role": "system"})
        messages.append({"content": prompt, "role": "user"})
        return messages

    @classmethod
    def prompt_to_input_tokens(
        cls,
        prompt: str,
        system_prompt: str = None,
        model: str = None,
    ) -> int:
        return cls.messages_to_input_tokens(
            cls.prompt_to_messages(prompt, system_prompt), model
        )

    @classmethod
    def get_input_tokens_instructor(
        cls,
        messages: str | list[dict[str, str]],
        model: str,
        response_model: BaseModel,
        **kwargs,  # just let people pass in whatever they want
    ):
        """
        From: https://community.openai.com/t/how-to-calculate-the-tokens-when-using-function-call/266573/11
        """
        instructor_messages = cls._get_raw_messages_instructor(
            messages=messages,
            model=model,
            response_model=response_model,
            process=True,
            **kwargs,
        )
        messages = instructor_messages["messages"]
        tools = instructor_messages["tools"]
        functions = [tool["function"] for tool in tools if tool["type"] == "function"]

        functions_tokens = 0
        functions_tokens += 12
        for function in functions:
            function_tokens = cls.tokenize(
                function["name"], model
            ) + cls.tokenize(function["description"], model)
            if "parameters" in function:
                parameters = function["parameters"]
                if "properties" in parameters:
                    function_tokens += 11
                    properties = parameters["properties"]
                    for properties_key, properties_value in properties.items():
                        function_tokens += cls.tokenize(
                            properties_key, model
                        )
                        for field, field_value in properties_value.items():
                            if field == "type" or field == "description":
                                function_tokens += (
                                    cls.tokenize(field_value, model) + 2
                                )
                            elif field == "enum":
                                function_tokens -= 3
                                for o in field_value:
                                    function_tokens += (
                                        cls.tokenize(o, model) + 3
                                    )
                            else:
                                warnings.warn(
                                    f"Field {field} not found in tokenization function"
                                )
            functions_tokens += function_tokens

        messages_tokens = cls.messages_to_input_tokens(messages, model)
        return messages_tokens + functions_tokens

    @classmethod
    def _get_raw_messages_instructor(
        cls,
        messages: str | list[dict[str, str]],
        model: str,
        response_model: BaseModel,
        process=True,
        **kwargs,  # just let people pass in whatever they want
    ) -> str | dict:
        
        import instructor
        from openai import OpenAI
        client = instructor.from_openai(OpenAI(api_key='MOCK'))
        
        if isinstance(messages, str):
            messages = [{"content": messages, "role": "user"}]
        log_stream = StringIO()
        logger = logging.getLogger("instructor")
        logger.setLevel(logging.DEBUG)
        stream_handler = logging.StreamHandler(log_stream)
        logger.addHandler(stream_handler)

        with patch("httpx.Client.send") as mock_send:
            mock_send.return_value = None  # prevent actual API calls

            try:
                bla = client.chat.completions.create(
                    messages=messages,
                    response_model=response_model,
                    model=model,
                )
            except Exception:
                # there WILL be an error
                # there's nothing you can do about it
                pass  # cope and seethe
        
        log_contents = log_stream.getvalue()
        for line in log_contents.splitlines():
            if "Instructor Request" in line:
                line = line.split("Instructor Request: ")[1]
                if process:
                    return cls._process_raw_prompt(line)
                return line

        warnings.warn(
            "No raw prompt found in logs. Maybe anthropic "
            "isn't supported or something idk"
        )
        return ""        

    @classmethod
    def _process_raw_prompt(cls, input_string: str) -> dict:
        # Step 1: Split at 'new_kwargs='
        split_parts = input_string.split("new_kwargs=", 1)
        if len(split_parts) < 2:
            warnings.warn(
                "Failed to split the string at 'new_kwargs='. Returning the original string."
            )
            return input_string

        dict_string = split_parts[1].strip()

        # Step 2: Attempt to load the remaining part as a dict
        dict_data = ast.literal_eval(dict_string)

        if not isinstance(dict_data, dict):
            warnings.warn(
                f"Decoded object is not a dictionary, but a {type(dict_data)}."
            )

        return dict_data

    """
    Cases:
    - Simulating
        - From input_tokens, output_tokens
        - From input_tokens, output_string
        - From input_tokens
        - From input_string
    - Real
        - From input_tokens, output_tokens, timer

    """

    @classmethod
    def _get_cost_simulating_from_input_tokens_output_tokens(
        cls,
        input_tokens: int,
        output_tokens_min: int,
        output_tokens_max: int,
        model: str,
        **kwargs,
    ) -> dict[str, float]:
        prices = cls.get_prices(model)
        cost_input_tokens = input_tokens * prices["input_tokens"]
        cost_output_tokens_min = output_tokens_min * prices["output_tokens"]
        cost_output_tokens_max = output_tokens_max * prices["output_tokens"]
        time_min = output_tokens_min * prices["time"]
        time_max = output_tokens_max * prices["time"]
        return {
            "cost_min": cost_input_tokens + cost_output_tokens_min,
            "cost_max": cost_input_tokens + cost_output_tokens_max,
            "time_min": time_min,
            "time_max": time_max,
            "input_tokens": input_tokens,
            "output_tokens_min": output_tokens_min,
            "output_tokens_max": output_tokens_max,
            "calls": 1,
            "model": model,
            "simulated": True,
            **kwargs,
        }

    @classmethod
    def _get_tokens(
        cls,
        model: str,
        input_tokens: int = None,
        output_tokens_min: int = None,
        output_tokens_max: int = None,
        input_string: str = None,
        messages: list[dict[str, str]] = None,
        output_string: str = None,
    ) -> tuple[int, int, int]:
        if input_tokens is None:
            try:
                assert input_string is not None or messages is not None
                if input_string is not None:
                    input_tokens = cls.tokenize(input_string, model)
                else:
                    input_tokens = cls.messages_to_input_tokens(
                        messages, model
                    )
            except:
                raise ValueError(
                    f"Failed to tokenize input_string {input_string} or messages {messages}"
                )
        if output_tokens_min is None or output_tokens_max is None:
            try:
                assert output_string is not None
                output_tokens = cls.tokenize(output_string, model)
                output_tokens_min, output_tokens_max = output_tokens, output_tokens
            except:
                try:
                    output_tokens_min, output_tokens_max = (
                        cls.output_tokens_estimate(
                            input_string=input_string,
                            messages=messages,
                            input_tokens=input_tokens,
                            model=model,
                        )
                    )
                except:
                    raise ValueError(
                        f"Failed to estimate output tokens for input_string "
                        f"{input_string} and input_tokens {input_tokens}"
                    )
        return input_tokens, output_tokens_min, output_tokens_max

    @classmethod
    def get_cost_simulating(
        cls,
        model: str,
        input_tokens: int = None,
        output_tokens_min: int = None,
        output_tokens_max: int = None,
        input_string: str = None,
        messages: list[dict[str, str]] = None,
        output_string: str = None,
        **kwargs,
    ) -> dict[str, float]:
        input_tokens, output_tokens_min, output_tokens_max = (
            cls._get_tokens(
                model=model,
                input_tokens=input_tokens,
                output_tokens_min=output_tokens_min,
                output_tokens_max=output_tokens_max,
                input_string=input_string,
                messages=messages,
                output_string=output_string,
            )
        )
        return cls._get_cost_simulating_from_input_tokens_output_tokens(
            input_tokens=input_tokens,
            output_tokens_min=output_tokens_min,
            output_tokens_max=output_tokens_max,
            model=model,
            input_string=input_string,
            messages=messages,
            output_string=output_string,
            **kwargs,
        )

    @classmethod
    def _get_cost_real_from_input_tokens_output_tokens_timer(
        cls,
        input_tokens: int, output_tokens: int, timer: float, model: str, **kwargs
    ) -> dict[str, float]:
        prices = cls.get_prices(model)
        cost_input_tokens = input_tokens * prices["input_tokens"]
        cost_output_tokens = output_tokens * prices["output_tokens"]
        time = timer
        return {
            "cost_min": cost_input_tokens + cost_output_tokens,
            "cost_max": cost_input_tokens + cost_output_tokens,
            "time_min": time,
            "time_max": time,
            "input_tokens": input_tokens,
            "output_tokens": output_tokens,
            "output_tokens_min": output_tokens,
            "output_tokens_max": output_tokens,
            "calls": 1,
            "model": model,
            "simulated": False,
            **kwargs,
        }

    @classmethod
    def get_cost_real(
        cls,
        model: str,
        input_tokens: int = None,
        output_tokens: int = None,
        input_string: str = None,
        messages: list[dict[str, str]] = None,
        output_string: str = None,
        timer: float = None,
        **kwargs,
    ) -> dict[str, float]:
        assert timer is not None
        input_tokens, output_tokens, _ = cls._get_tokens(
            model=model,
            input_tokens=input_tokens,
            output_tokens_min=output_tokens,
            output_tokens_max=output_tokens,
            input_string=input_string,
            messages=messages,
            output_string=output_string,
        )

        return cls._get_cost_real_from_input_tokens_output_tokens_timer(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            timer=timer,
            model=model,
            input_string=input_string,
            messages=messages,
            output_string=output_string,
            **kwargs,
        )



import os
import datetime
import jsonlines
import time
import asyncio
from dataclasses import dataclass
from typing import Literal
from contextlib import contextmanager, asynccontextmanager
from pathlib import Path


class Costlog:

    def __init__(
        self,
        path: Path | str = None,
        mode: Literal["jsonl", "memory"] = "memory",
        totals_keys: set[str] = None,
        overwrite: bool = None,
    ):
        """
        Arguments:
            path: Path to the costlog file. If None, the costlog will be stored in
                the current working directory in a folder called ".costly".
            mode: "jsonl" or "memory". "jsonl" will store the costlog in a jsonl file,
                "memory" will store the costlog in memory which may not be suitable for large logs.
            totals_keys: The keys to store in the totals.
            overwrite: If the costlog file already exists, overwrite it. If None, the user will be asked. Only relevant if mode is "jsonl".
        """
        match mode:
            case "jsonl":
                if path is None:
                    self.path = Path(
                        os.getcwd(),
                        ".costly",
                        f"costlog_{datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')}.jsonl",
                    )
                else:
                    self.path = Path(path)
                if self.path.exists():
                    if overwrite is None:
                        overwrite = input(f"Costlog file {self.path} already exists. Overwrite? (y/n) ")
                    if overwrite not in ["y", True]:
                        raise FileExistsError(f"Costlog file {self.path} already exists.")
                    else:
                        self.path.unlink()
                else:
                    self.path.parent.mkdir(parents=True, exist_ok=True)
                    with open(self.path, "w") as f:
                        jsonlines.Writer(f)
            case "memory":
                self.items = []
            case _:
                raise ValueError(f"Invalid mode: {mode}")
        self.mode = mode
        if totals_keys is None:
            totals_keys = {"cost_min", "cost_max", "time_min", "time_max", "calls"}
        self.totals_keys = totals_keys
        self.totals = {key: 0.0 for key in totals_keys}

    def append(self, **kwargs):
        match self.mode:
            case "jsonl":
                with jsonlines.open(self.path, "a") as f:
                    f.write(kwargs)
            case "memory":
                self.items.append(kwargs)
        for key in self.totals:
            self.totals[key] += kwargs.get(key, 0.0)

    @contextmanager
    def new_item(self):
        item = {}
        t1 = time.perf_counter()
        yield item, lambda: time.perf_counter() - t1
        self.append(**item)

    @asynccontextmanager
    async def new_item_async(self):
        item = {}
        t1 = time.perf_counter()
        yield item, lambda: time.perf_counter() - t1
        self.append(**item)

    def totals_from_items(self):
        match self.mode:
            case "jsonl":
                with jsonlines.open(self.path) as f:
                    items = list(f)
            case "memory":
                items = self.items
            case _:
                raise ValueError(f"Invalid mode: {self.mode}")
        return {key: sum(item[key] for item in items) for key in self.totals_keys}
